"""
This reads in the data from files.  It makes spike generators to
turn on the first layer, and associated generators to turn on the
second layer shortly thereafter, so it can learn by STDP.
The layers are connected by STDP synapses.
At the end of the run, it tests by using spike sources to turn on
the input layer, and the test is the output layer.
"""
import sys
import cPickle as pickle

import pyNN.nest as sim

#EXAMPLE_TIME = 30.0
EXAMPLE_TIME = 50.0
#NUMBER_TRAINING_EPOCHS = 5
NUMBER_TRAINING_EPOCHS = 6
#NUMBER_TRAINING_EPOCHS = 7
if (EXAMPLE_TIME == 30.0):
    RUN_DURATION = 15000 #30
    END_OF_TRAINING = 12000 
elif ((EXAMPLE_TIME == 50.0) and (NUMBER_TRAINING_EPOCHS == 6)):
    END_OF_TRAINING = 23000 
    RUN_DURATION = 28000
elif ((EXAMPLE_TIME == 50.0) and (NUMBER_TRAINING_EPOCHS == 7)):
    END_OF_TRAINING = 26000 
    RUN_DURATION = 30000
elif (EXAMPLE_TIME == 50.0):
    RUN_DURATION = 25000
    END_OF_TRAINING = 19000 


TIME_STEP = 1.0
FIRING_WEIGHT = 0.1

neuronsInFirstLayer = 404
neuronsInSecondLayer = 3
inputFeatureBreadth = 3

firstSecond = -1 #global for printing out synaptic weights

def readData(fileName):
    data = []
    dataFileHandle = open(fileName,'r')
    for lineNumber in range (0,75):
        inputLine = dataFileHandle.readline()
        dataItemStrings = inputLine.split()
        feature0 = float(dataItemStrings[0])
        feature0 = int(feature0*100)
        feature1 = float(dataItemStrings[1])
        feature1 = int(feature1*100)
        feature2 = float(dataItemStrings[2])
        feature2 = int(feature2*100)
        feature3 = float(dataItemStrings[3])
        feature3 = int(feature3*100)
        outputFeature = int(dataItemStrings[4])
        dataLine = [feature0,feature1,feature2,feature3,outputFeature]
        data = data + [dataLine]

    dataFileHandle.close()
    return data
    

#initialize the simulator. 
def init(): 
    sim.setup(timestep=TIME_STEP,min_delay=TIME_STEP,
                    max_delay=TIME_STEP, debug=0)

#Make spike sources to turn on the first layer, and the second layer for
#training.
def createTrainSources(data):
    numberDataItems = len(data)

    start = 10.0
    learnOffset = 1
    epochTime = EXAMPLE_TIME*numberDataItems
    spikeSources = []
    inputSpikeTimesMatrix = []
    outputSpikeTimesMatrix = []
    for dataOffset in range (0,numberDataItems):
        #print data[dataOffset]
        inputSpikeTimes = []
        outputSpikeTimes = []

        for epoch in range (0,NUMBER_TRAINING_EPOCHS): 
            inputSpikeTime = start+(dataOffset*EXAMPLE_TIME)+(epoch*epochTime)
            inputSpikeTimes = inputSpikeTimes + [inputSpikeTime]
            outputSpikeTime = inputSpikeTime + learnOffset
            outputSpikeTimes = outputSpikeTimes + [outputSpikeTime]
        inputSpikeTimesMatrix = inputSpikeTimesMatrix + [inputSpikeTimes]
        outputSpikeTimesMatrix = outputSpikeTimesMatrix + [outputSpikeTimes]


    inputSpikeGens = sim.Population(numberDataItems,sim.SpikeSourceArray,
                {'spike_times': inputSpikeTimesMatrix},label="inputSpikes")

    outputSpikeGens = sim.Population(numberDataItems,sim.SpikeSourceArray,
                {'spike_times': outputSpikeTimesMatrix},label="outputSpikes")

    return [inputSpikeGens,outputSpikeGens]

def createTestSources(data,startTime):
    numberDataItems = len(data)
    timeBetweenTests = EXAMPLE_TIME
    spikeSources = []
    testSpikeTimesMatrix = []
    for dataOffset in range (0,numberDataItems):
        testSpikeTimes = [[startTime + (dataOffset*timeBetweenTests)]]
        testSpikeTimesMatrix = testSpikeTimesMatrix + testSpikeTimes

    testSpikeGens = sim.Population(numberDataItems,sim.SpikeSourceArray,
                {'spike_times': testSpikeTimesMatrix},label="testSpikes")

    return testSpikeGens

def convertDataItemToSynapseDescription(dataLine,sourceNumber):
    inputConnector = []
    neuronsPerInputFeature = 101
    for dataFeature in range (0,4):
        dataFeatureValue = dataLine[dataFeature]
        for neuronOffset in range((dataFeatureValue - inputFeatureBreadth),
                              (dataFeatureValue + inputFeatureBreadth+1)):
            if ((neuronOffset >= 0) and (neuronOffset<neuronsPerInputFeature)):
                toNeuron = neuronOffset+(dataFeature*neuronsPerInputFeature)
                inputConnector = inputConnector + [(sourceNumber,toNeuron,
                        FIRING_WEIGHT,TIME_STEP)]

    return inputConnector

def connectInput(spikeSources, data,inputCells,outputCells):
    inputSpikeSources = spikeSources[0]
    outputSpikeSources = spikeSources[1]
    numberDataItems = len(data)
    for dataOffset in range (0,numberDataItems):
        dataItem = data[dataOffset]
        inputConnector=convertDataItemToSynapseDescription(dataItem,dataOffset)
        fromSource = dataOffset
        inputFromListConnector = sim.FromListConnector(inputConnector)
        sim.Projection(inputSpikeSources, inputCells, inputFromListConnector, 
                   receptor_type='excitatory')

        outputCategory = dataItem[4]-1
        outputConnector = [(fromSource,outputCategory,FIRING_WEIGHT,
                            TIME_STEP)]
            
        outputFromListConnector = sim.FromListConnector(outputConnector)
        sim.Projection(outputSpikeSources, outputCells,outputFromListConnector, 
                   receptor_type='excitatory')

def connectTest(testSpikeSources, data,inputCells):
    numberDataItems = len(data)
    for dataOffset in range (0,numberDataItems):
        dataItem = data[dataOffset]
        testConnector=convertDataItemToSynapseDescription(dataItem,dataOffset)
        testFromListConnector = sim.FromListConnector(testConnector)
        sim.Projection(testSpikeSources, inputCells, testFromListConnector, 
                   receptor_type='excitatory')

def connectFirstToSecondLayer(fromCells, toCells):
    global firstSecond
    connector = []
    stdp_model = sim.STDPMechanism(
        timing_dependence=sim.SpikePairRule(tau_plus=20.0,tau_minus=20.0,
                                            A_plus=0.004, A_minus=0.003),
        weight_dependence=sim.AdditiveWeightDependence(w_min=0.0, w_max=0.03),
        weight=0.0000,
        delay=1.0,
        dendritic_delay_fraction=float(1))

    firstSecond = sim.Projection(fromCells,toCells,
                   sim.AllToAllConnector(),
                   synapse_type = stdp_model)

def createDefaultNeurons(numberNeurons):
    cells=sim.Population(numberNeurons,sim.IF_cond_exp,cellparams={})
    return cells

def setupRecording(cells):
    cells.record({'spikes','v'})

def printResults(simCells,file):
    simCells.write_data(file+".pkl",'spikes')
    #simCells.write_data(file+"V.pkl",'v')
    #simCells.write_data(file+"W.pkl",'w')

##--main---
args = sys.argv
numberArgs = args.__len__()
if (numberArgs == 3):
    trainFileName = args[1]
    testFileName = args[2]
else:
    trainFileName = "irisTrain.txt"
    testFileName = "irisTest.txt"

print trainFileName, testFileName


simName = "nest"
spinnVersion = -1

trainData = readData(trainFileName)
testData = readData(testFileName)

init()
trainSpikeGenerators = createTrainSources(trainData)
testSpikeGenerators = createTestSources(testData,END_OF_TRAINING)

firstLayerCells = createDefaultNeurons(neuronsInFirstLayer)
secondLayerCells = createDefaultNeurons(neuronsInSecondLayer)

#hook it up
connectInput(trainSpikeGenerators,trainData,firstLayerCells,secondLayerCells)
connectTest(testSpikeGenerators,testData,firstLayerCells)
connectFirstToSecondLayer(firstLayerCells,secondLayerCells)

setupRecording(firstLayerCells)
setupRecording(secondLayerCells)

sim.run(RUN_DURATION)

printResults(firstLayerCells,"output/first")
printResults(secondLayerCells,"output/second")

#firstSecondSynapseWeights = firstSecond.get(["weight"],format="list")
#print firstSecondSynapseWeights

sim.reset()
