"""
This reads in the data from files.  It makes spike generators to
turn on the first layer, and associated generators to turn on the
second layers shortly thereafter, so it can learn by STDP.
The layers are connected by STDP synapses.

At the end of the run, it tests by using spike sources to turn on
the input layer, and the test is the output layer.
"""
import sys
import cPickle as pickle

import pyNN.nest as sim

RUN_DURATION = 15000
END_OF_TRAINING = 12000
TIME_STEP = 1.0
FIRING_WEIGHT = 0.1

neuronsPerInputFeature = 101
neuronsInFirstLayer = 404
neuronsInSecondLayer = 3
neuronsInThirdLayer = 3
inputFeatureBreadth = 3
numberCategories = 3

firstSecond = -1

def readData(fileName):
    data = []
    dataFileHandle = open(fileName,'r')
    for lineNumber in range (0,75):
        inputLine = dataFileHandle.readline()
        dataItemStrings = inputLine.split()
        feature0 = float(dataItemStrings[0])
        feature0 = int(feature0*100)
        feature1 = float(dataItemStrings[1])
        feature1 = int(feature1*100)
        feature2 = float(dataItemStrings[2])
        feature2 = int(feature2*100)
        feature3 = float(dataItemStrings[3])
        feature3 = int(feature3*100)
        outputFeature = int(dataItemStrings[4])
        dataLine = [feature0,feature1,feature2,feature3,outputFeature]
        data = data + [dataLine]

    dataFileHandle.close()
    return data
    

#initialize the simulator. 
def init(): 
    sim.setup(timestep=TIME_STEP,min_delay=TIME_STEP,
                    max_delay=TIME_STEP, debug=0)

#Make spike sources to turn on the first layer, and the second layer for
#training.
def createTrainSources(data):
    numberDataItems = len(data)

    start = 10.0
    learnOffset = 1
    exampleTime = 30
    epochTime = exampleTime*numberDataItems
    numberTrainingEpochs = 5
    spikeSources = []
    inputSpikeTimesMatrix = []
    outputSpikeTimesMatrix = []
    for dataOffset in range (0,numberDataItems):
        #print data[dataOffset]
        inputSpikeTimes = []
        outputSpikeTimes = []

        for epoch in range (0,numberTrainingEpochs): 
            inputSpikeTime = start+(dataOffset*exampleTime)+(epoch*epochTime)
            inputSpikeTimes = inputSpikeTimes + [inputSpikeTime]
            outputSpikeTime = inputSpikeTime + learnOffset
            outputSpikeTimes = outputSpikeTimes + [outputSpikeTime]
        inputSpikeTimesMatrix = inputSpikeTimesMatrix + [inputSpikeTimes]
        outputSpikeTimesMatrix = outputSpikeTimesMatrix + [outputSpikeTimes]


    inputSpikeGens = sim.Population(numberDataItems,sim.SpikeSourceArray,
                {'spike_times': inputSpikeTimesMatrix},label="inputSpikes")

    outputSpikeGens = sim.Population(numberDataItems,sim.SpikeSourceArray,
                {'spike_times': outputSpikeTimesMatrix},label="outputSpikes")

    return [inputSpikeGens,outputSpikeGens]

def createTestSources(data,startTime):
    numberDataItems = len(data)
    timeBetweenTests = 30
    spikeSources = []
    testSpikeTimesMatrix = []
    for dataOffset in range (0,numberDataItems):
        testSpikeTimes = [[startTime + (dataOffset*timeBetweenTests)]]
        testSpikeTimesMatrix = testSpikeTimesMatrix + testSpikeTimes

    testSpikeGens = sim.Population(numberDataItems,sim.SpikeSourceArray,
                {'spike_times': testSpikeTimesMatrix},label="testSpikes")

    return testSpikeGens

def convertDataItemToSynapseDescription(dataLine,sourceNumber):
    inputConnector = []
    for dataFeature in range (0,4):
        dataFeatureValue = dataLine[dataFeature]
        for neuronOffset in range((dataFeatureValue - inputFeatureBreadth),
                              (dataFeatureValue + inputFeatureBreadth+1)):
            if ((neuronOffset >= 0) and (neuronOffset<neuronsPerInputFeature)):
                toNeuron = neuronOffset+(dataFeature*neuronsPerInputFeature)
                inputConnector = inputConnector + [(sourceNumber,toNeuron,
                        FIRING_WEIGHT,TIME_STEP)]

    return inputConnector

def connectInput(spikeSources, data,inputCells):
    inputSpikeSources = spikeSources[0]
    numberDataItems = len(data)
    for dataOffset in range (0,numberDataItems):
        dataItem = data[dataOffset]
        inputConnector=convertDataItemToSynapseDescription(dataItem,dataOffset)
        fromSource = dataOffset
        inputFromListConnector = sim.FromListConnector(inputConnector)
        sim.Projection(inputSpikeSources, inputCells, inputFromListConnector, 
                   receptor_type='excitatory')

def connectTrainOutput(spikeSources, data,outputCells):
    outputSpikeSources = spikeSources[1]
    numberDataItems = len(data)
    for dataOffset in range (0,numberDataItems):
        dataItem = data[dataOffset]
        outputCategory = dataItem[4]-1
        outputSpikeSource = dataOffset
        outputConnector = [(outputSpikeSource,outputCategory,FIRING_WEIGHT,
                            TIME_STEP)]
            
        outputFromListConnector = sim.FromListConnector(outputConnector)
        sim.Projection(outputSpikeSources, outputCells,outputFromListConnector, 
                   receptor_type='excitatory')

def connectTest(testSpikeSources, data,inputCells):
    numberDataItems = len(data)
    spikeSourceOffset = 0
    for dataOffset in range (0,numberDataItems):
        dataItem = data[dataOffset]
        testConnector=convertDataItemToSynapseDescription(dataItem,dataOffset)
        testFromListConnector = sim.FromListConnector(testConnector)
        sim.Projection(testSpikeSources, inputCells, testFromListConnector, 
                   receptor_type='excitatory')

def connectFirstToSecondLayer(fromCells, toCells):
    global firstSecond

    connector = []
    stdp_model = sim.STDPMechanism(
        timing_dependence=sim.SpikePairRule(tau_plus=20.0,tau_minus=20.0,
                                            A_plus=0.004, A_minus=0.003),
        weight_dependence=sim.AdditiveWeightDependence(w_min=0, w_max=0.05),
        weight=0.0,
        delay=1.0,
        dendritic_delay_fraction=float(1))

    firstSecond = sim.Projection(fromCells,toCells,
                   sim.AllToAllConnector(),
                   synapse_type = stdp_model)

def getFromFeatureConnector(inputFeatureNumber):
    connector = []
    for outputNeuron in range (0,numberCategories):
        for inputNeuronOffset in range (0,neuronsPerInputFeature):
            inputNeuron = (inputFeatureNumber*neuronsPerInputFeature + 
                           inputNeuronOffset)
            connector = connector + [(inputNeuron,outputNeuron,0.0,TIME_STEP)]

    return connector
    
def connectPartialFirstToSecondLayer(fromCells, toCells,inputFeatureA,
                                     inputFeatureB):
    stdp_model = sim.STDPMechanism(
        timing_dependence=sim.SpikePairRule(tau_plus=20.0,tau_minus=20.0,
                                            A_plus=0.004, A_minus=0.003),
        weight_dependence=sim.AdditiveWeightDependence(w_min=0, w_max=0.05),
        weight=0.0,
        delay=1.0,
        dendritic_delay_fraction=float(1))

    connector = getFromFeatureConnector(inputFeatureA)
    connector = connector + getFromFeatureConnector(inputFeatureB)

    #hook it up
    fromListConnector = sim.FromListConnector(connector)
    sim.Projection(fromCells,toCells,
                   fromListConnector,
                   synapse_type = stdp_model)

#it's all one to one by category, so make a straight static weight
def connectSecondToThird(secondCells,thirdCells,weight):
    connector = []
    for neuron in range (0,numberCategories):
        connector = connector + [(neuron,neuron,weight,TIME_STEP)]

    fromListConnector = sim.FromListConnector(connector)
    sim.Projection(secondCells,thirdCells,fromListConnector,
                   receptor_type='excitatory')


def createDefaultNeurons(numberNeurons):
    cells=sim.Population(numberNeurons,sim.IF_cond_exp,cellparams={})
    return cells

def setupRecording(cells):
    cells.record({'spikes','v'})

def printResults(simCells,file):
    simCells.write_data(file+".pkl",'spikes')
    #simCells.write_data(file+"V.pkl",'v')
    #simCells.write_data(file+"W.pkl",'w')

##--main---
args = sys.argv
numberArgs = args.__len__()
print args, numberArgs
if (numberArgs == 5):
    trainFile = args[1]
    testFile = args[2]
    allWt = int(args[3])
    allWt = allWt/100.0
    featureWt = int (args[4])
    featureWt = featureWt/100.0
    print "Test", allWt, featureWt
elif (numberArgs == 3):
    trainFile = args[1]
    testFile = args[2]
    allWt = 0.06
    featureWt = 0.02
else:
    trainFile = "irisTrain.txt"
    testFile = "irisTest.txt"
    allWt = 0.06
    featureWt = 0.02

print "Test2", allWt, featureWt

simName = "nest"
spinnVersion = -1

trainData = readData(trainFile)
testData = readData(testFile)

init()
trainSpikeGenerators = createTrainSources(trainData)
testSpikeGenerators = createTestSources(testData,12000)

firstLayerCells = createDefaultNeurons(neuronsInFirstLayer)
secondLayerPrimaryCells = createDefaultNeurons(neuronsInSecondLayer)

secondLayerOneTwoCells = createDefaultNeurons(neuronsInSecondLayer)
secondLayerOneThreeCells = createDefaultNeurons(neuronsInSecondLayer)
secondLayerOneFourCells = createDefaultNeurons(neuronsInSecondLayer)
secondLayerTwoThreeCells = createDefaultNeurons(neuronsInSecondLayer)
secondLayerTwoFourCells = createDefaultNeurons(neuronsInSecondLayer)
secondLayerThreeFourCells = createDefaultNeurons(neuronsInSecondLayer)

thirdLayerCells = createDefaultNeurons(neuronsInThirdLayer)

setupRecording(firstLayerCells)
setupRecording(secondLayerPrimaryCells)
setupRecording(secondLayerOneTwoCells)
setupRecording(secondLayerOneThreeCells)
setupRecording(secondLayerOneFourCells)
setupRecording(secondLayerTwoThreeCells)
setupRecording(secondLayerTwoFourCells)
setupRecording(secondLayerThreeFourCells)
setupRecording(thirdLayerCells)

#hook it up
connectInput(trainSpikeGenerators,trainData,firstLayerCells)
connectTrainOutput(trainSpikeGenerators,trainData,secondLayerPrimaryCells)
connectTrainOutput(trainSpikeGenerators,trainData,secondLayerOneTwoCells)
connectTrainOutput(trainSpikeGenerators,trainData,secondLayerOneThreeCells)
connectTrainOutput(trainSpikeGenerators,trainData,secondLayerOneFourCells)
connectTrainOutput(trainSpikeGenerators,trainData,secondLayerTwoThreeCells)
connectTrainOutput(trainSpikeGenerators,trainData,secondLayerTwoFourCells)
connectTrainOutput(trainSpikeGenerators,trainData,secondLayerThreeFourCells)

connectTest(testSpikeGenerators,testData,firstLayerCells)

connectFirstToSecondLayer(firstLayerCells,secondLayerPrimaryCells)
connectPartialFirstToSecondLayer(firstLayerCells,secondLayerOneTwoCells,0,1)
connectPartialFirstToSecondLayer(firstLayerCells,secondLayerOneThreeCells,0,2)
connectPartialFirstToSecondLayer(firstLayerCells,secondLayerOneFourCells,0,3)
connectPartialFirstToSecondLayer(firstLayerCells,secondLayerTwoThreeCells,1,2)
connectPartialFirstToSecondLayer(firstLayerCells,secondLayerTwoFourCells,1,3)
connectPartialFirstToSecondLayer(firstLayerCells,secondLayerThreeFourCells,2,3)

connectSecondToThird(secondLayerPrimaryCells,thirdLayerCells,allWt)
connectSecondToThird(secondLayerOneFourCells,thirdLayerCells,featureWt)
connectSecondToThird(secondLayerOneTwoCells,thirdLayerCells,featureWt)
connectSecondToThird(secondLayerOneThreeCells,thirdLayerCells,featureWt)
connectSecondToThird(secondLayerTwoThreeCells,thirdLayerCells,featureWt)
connectSecondToThird(secondLayerTwoFourCells,thirdLayerCells,featureWt)
connectSecondToThird(secondLayerThreeFourCells,thirdLayerCells,featureWt)

sim.run(RUN_DURATION)

printResults(firstLayerCells,"output/input")
printResults(secondLayerPrimaryCells,"output/secondAll")
printResults(secondLayerOneTwoCells,"output/secondOneTwo")
printResults(secondLayerOneThreeCells,"output/secondOneThree")
printResults(secondLayerOneFourCells,"output/secondOneFour")
printResults(secondLayerTwoThreeCells,"output/secondTwoThree")
printResults(secondLayerTwoFourCells,"output/secondTwoFour")
printResults(secondLayerThreeFourCells,"output/secondThreeFour")
printResults(thirdLayerCells,"output/third")

firstSecondSynapseWeights = firstSecond.get(["weight"],format="list")
#print firstSecondSynapseWeights
sim.reset()
